/*
    Project: TCTEMP
    Version: V1.5
    File   : tc_password.c

    Copyright (c) 2006-2007 Author of TCTEMP. All rights reserved.
    Public Key Fingerprint: 75EB 6BC2 01B7 F6E7 4BD7 CC58 4A5F C393 19EE 6E69

    Covered by TrueCrypt License 2.2 the full text of which is contained
    in the file License.txt included in TCTEMP binary and source code
    distribution archives.

    The copyright notice of this source code file and the public key
    fingerprint must be retained in full.
*/

#include "tc_priv.h"

//#define USE_ENTROPY_ESTIMATION

#ifndef USE_ENTROPY_ESTIMATION

/*----------------------------------------------------------------------------*/
/* Simple password strength test                                              */

#define MIN_DIFF_PASSWORD_CHAR_COUNT 10

static unsigned __stdcall TC_NumOfDiffPasswordChars(PTC_PASSWORD
  pPassword, PTC_PASSWORD_INFO pPasswordInfo)
{
    const TC_UINT8 *pb;
    const TC_UINT8 *pbEnd;
    unsigned       nChars;

    if (pPassword->cbPasswordSize == 0) {
        return 0;
    }

    memset(pPasswordInfo->rgbFrequency, '\0',
      sizeof(pPasswordInfo->rgbFrequency));
    pb = pPassword->rgbPassword;
    pbEnd = pb + pPassword->cbPasswordSize;
    for (; pb != pbEnd; pb++) {
        pPasswordInfo->rgbFrequency[*pb]++;
    }

    pb = pPasswordInfo->rgbFrequency;
    pbEnd = pb + 256;
    for (nChars = 0; pb != pbEnd; pb++) {
        if (*pb) {
            nChars++;
        }
    }

    return nChars;
}

BOOL __stdcall TC_PasswordIsWeak(PTC_PASSWORD pPassword,
  unsigned uDesiredStrength, PTC_PASSWORD_INFO pPasswordInfo)
{
    const TC_UINT8 *pbSrc;
    const TC_UINT8 *pbSrcEnd;
    TC_UINT8       *pbDest;

    UNREFERENCED_PARAMETER(uDesiredStrength);

    if (TC_NumOfDiffPasswordChars(pPassword, pPasswordInfo) <
      MIN_DIFF_PASSWORD_CHAR_COUNT) {
        return TRUE;
    }

    pPasswordInfo->pwTemp.cbPasswordSize = pPassword->cbPasswordSize - 1;
    pbSrc = pPassword->rgbPassword;
    pbSrcEnd = pbSrc + pPassword->cbPasswordSize;
    pbDest = pPasswordInfo->pwTemp.rgbPassword;
    for (pbSrc++; pbSrc != pbSrcEnd; pbSrc++) {
        *pbDest++ = pbSrc[0] - pbSrc[-1];
    }

    if (TC_NumOfDiffPasswordChars(&pPasswordInfo->pwTemp, pPasswordInfo) <
      MIN_DIFF_PASSWORD_CHAR_COUNT) {
        return TRUE;
    }

    return FALSE;
}

#else

/*----------------------------------------------------------------------------*/
/* Password strength test with entropy estimation                             */

#include <math.h>
#ifndef M_LOG2E
#  define M_LOG2E 1.44269504088896340736
#endif

static unsigned __stdcall TC_EstimateEffectivePasswordSize(
  PTC_PASSWORD pPassword, PTC_PASSWORD_INFO pPasswordInfo)
{
    const TC_UINT8 *pb;
    const TC_UINT8 *pbEnd;
    double         dBitSize;
    double         dInvCount;

    if (pPassword->cbPasswordSize == 0) {
        return 0;
    }
    dInvCount = 1. / pPassword->cbPasswordSize;

    memset(pPasswordInfo->rgbFrequency, '\0',
      sizeof(pPasswordInfo->rgbFrequency));
    pb = pPassword->rgbPassword;
    pbEnd = pb + pPassword->cbPasswordSize;
    for (; pb != pbEnd; pb++) {
        pPasswordInfo->rgbFrequency[*pb]++;
    }

    pb = pPasswordInfo->rgbFrequency;
    pbEnd = pb + 256;
    for (dBitSize = 0; pb != pbEnd; pb++) {
        if (*pb) {
            dBitSize -= *pb * log(*pb * dInvCount);
        }
    }
    if (dBitSize < 0) {
        return 0;
    }

    return (unsigned)(dBitSize * M_LOG2E);
}

static unsigned __stdcall TC_EstimatePasswordStrength(PTC_PASSWORD pPassword,
  PTC_PASSWORD_INFO pPasswordInfo)
{
    const TC_UINT8 *pbSrc;
    const TC_UINT8 *pbSrcEnd;
    TC_UINT8       *pbDest;
    unsigned       uBitSize1;
    unsigned       uBitSize2;

    uBitSize1 = TC_EstimateEffectivePasswordSize(pPassword, pPasswordInfo);
    if (uBitSize1 == 0) {
        return 0;
    }

    pPasswordInfo->pwTemp.cbPasswordSize = pPassword->cbPasswordSize - 1;
    pbSrc = pPassword->rgbPassword;
    pbSrcEnd = pbSrc + pPassword->cbPasswordSize;
    pbDest = pPasswordInfo->pwTemp.rgbPassword;
    for (pbSrc++; pbSrc != pbSrcEnd; pbSrc++) {
        *pbDest++ = pbSrc[0] - pbSrc[-1];
    }

    uBitSize2 = TC_EstimateEffectivePasswordSize(&pPasswordInfo->pwTemp,
      pPasswordInfo);

    return (uBitSize1 < uBitSize2) ? uBitSize1 : uBitSize2;
}

BOOL __stdcall TC_PasswordIsWeak(PTC_PASSWORD pPassword,
  unsigned uDesiredStrength, PTC_PASSWORD_INFO pPasswordInfo)
{
    return TC_EstimatePasswordStrength(pPassword, pPasswordInfo) <
      uDesiredStrength;
}

#endif